__all__ = [
    'ODENet'
]

import torch
from .velocity_net import VelocityNet

class ODENet(torch.nn.Module):

    def __init__(self, n_frames : int = 10, **vnet_kwargs):
        super().__init__()
        self.n_frames = n_frames
        self.v = VelocityNet(**vnet_kwargs)

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        for _ in range(self.n_frames):
            dx = self.v(x) / self.n_frames
            x = x + dx
        return x

    def trajectory(self, x : torch.Tensor):
        yield x
        for _ in range(self.n_frames):
            dx = self.v(x) / self.n_frames
            x = x + dx
            yield x



__all__ = [
    'Gridding'
]

import cupy as cp
import torch
from typing import Tuple
from .utility import torch_to_cupy, cupy_to_torch, BLOCKDIM, BLOCKSIZE

ker_gridding3d = cp.RawKernel(r'''
extern "C" __global__ void gridding3d(
    const float* rho,
    const float* xs,
    int m,
    float* vol,
    int n1,
    int n2,
    int n3,
    float pixel_size)
{
    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    if (tid < m) {
        float x_ = (xs[tid * 3    ] - pixel_size / 2) / pixel_size;
        float y_ = (xs[tid * 3 + 1] - pixel_size / 2) / pixel_size;
        float z_ = (xs[tid * 3 + 2] - pixel_size / 2) / pixel_size;
        float wt = rho[tid];
        int x = floorf(x_);
        int y = floorf(y_);
        int z = floorf(z_);
        float dx = x_ - x;
        float dy = y_ - y;
        float dz = z_ - z;
        if (0 <= x     && x     < n1 && 0 <= y     && y     < n2 && 0 <= z     && z     < n3) atomicAdd(vol + ( x      * n2 + y    ) * n3 + z    , wt * (1 - dx) * (1 - dy) * (1 - dz));
        if (0 <= x + 1 && x + 1 < n1 && 0 <= y     && y     < n2 && 0 <= z     && z     < n3) atomicAdd(vol + ((x + 1) * n2 + y    ) * n3 + z    , wt *      dx  * (1 - dy) * (1 - dz));
        if (0 <= x     && x     < n1 && 0 <= y + 1 && y + 1 < n2 && 0 <= z     && z     < n3) atomicAdd(vol + ( x      * n2 + y + 1) * n3 + z    , wt * (1 - dx) *      dy  * (1 - dz));
        if (0 <= x + 1 && x + 1 < n1 && 0 <= y + 1 && y + 1 < n2 && 0 <= z     && z     < n3) atomicAdd(vol + ((x + 1) * n2 + y + 1) * n3 + z    , wt *      dx  *      dy  * (1 - dz));
        if (0 <= x     && x     < n1 && 0 <= y     && y     < n2 && 0 <= z + 1 && z + 1 < n3) atomicAdd(vol + ( x      * n2 + y    ) * n3 + z + 1, wt * (1 - dx) * (1 - dy) *      dz );
        if (0 <= x + 1 && x + 1 < n1 && 0 <= y     && y     < n2 && 0 <= z + 1 && z + 1 < n3) atomicAdd(vol + ((x + 1) * n2 + y    ) * n3 + z + 1, wt *      dx  * (1 - dy) *      dz );
        if (0 <= x     && x     < n1 && 0 <= y + 1 && y + 1 < n2 && 0 <= z + 1 && z + 1 < n3) atomicAdd(vol + ( x      * n2 + y + 1) * n3 + z + 1, wt * (1 - dx) *      dy  *      dz );
        if (0 <= x + 1 && x + 1 < n1 && 0 <= y + 1 && y + 1 < n2 && 0 <= z + 1 && z + 1 < n3) atomicAdd(vol + ((x + 1) * n2 + y + 1) * n3 + z + 1, wt *      dx  *      dy  *      dz );
    }
}
''', 'gridding3d')

ker_grad_gridding3d = cp.RawKernel(r'''
extern "C" __global__ void grad_gridding3d(
    const float* rho,
    const float* xs,
    int m,
    const float* dvol,
    int n1,
    int n2,
    int n3,
    float pixel_size,
    float* vs)
{
    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    if (tid < m) {
        float x_ = (xs[tid * 3    ] - pixel_size / 2) / pixel_size;
        float y_ = (xs[tid * 3 + 1] - pixel_size / 2) / pixel_size;
        float z_ = (xs[tid * 3 + 2] - pixel_size / 2) / pixel_size;
        float wt = rho[tid];
        int x = floorf(x_);
        int y = floorf(y_);
        int z = floorf(z_);
        float dx = x_ - x;
        float dy = y_ - y;
        float dz = z_ - z;
        float vx = 0;
        float vy = 0;
        float vz = 0;
        if (0 <= x     && x     < n1 && 0 <= y     && y     < n2 && 0 <= z     && z     < n3) {
            float grad = dvol[( x      * n2 + y    ) * n3 + z    ];
            vx += wt *    -  1  * (1 - dy) * (1 - dz) * grad;
            vy += wt * (1 - dx) *    -  1  * (1 - dz) * grad;
            vz += wt * (1 - dx) * (1 - dy) *    -  1  * grad;
        }
        if (0 <= x + 1 && x + 1 < n1 && 0 <= y     && y     < n2 && 0 <= z     && z     < n3) {
            float grad = dvol[((x + 1) * n2 + y    ) * n3 + z    ];
            vx += wt *       1  * (1 - dy) * (1 - dz) * grad;
            vy += wt *      dx  *    -  1  * (1 - dz) * grad;
            vz += wt *      dx  * (1 - dy) *    -  1  * grad;
        }
        if (0 <= x     && x     < n1 && 0 <= y + 1 && y + 1 < n2 && 0 <= z     && z     < n3) {
            float grad = dvol[( x      * n2 + y + 1) * n3 + z    ];
            vx += wt *    -  1  *      dy  * (1 - dz) * grad;
            vy += wt * (1 - dx) *       1  * (1 - dz) * grad;
            vz += wt * (1 - dx) *      dy  *    -  1  * grad;
        }
        if (0 <= x + 1 && x + 1 < n1 && 0 <= y + 1 && y + 1 < n2 && 0 <= z     && z     < n3) {
            float grad = dvol[((x + 1) * n2 + y + 1) * n3 + z    ];
            vx += wt *       1  *      dy  * (1 - dz) * grad;
            vy += wt *      dx  *       1  * (1 - dz) * grad;
            vz += wt *      dx  *      dy  *    -  1  * grad;
        }
        if (0 <= x     && x     < n1 && 0 <= y     && y     < n2 && 0 <= z + 1 && z + 1 < n3) {
            float grad = dvol[( x      * n2 + y    ) * n3 + z + 1];
            vx += wt *    -  1  * (1 - dy) *      dz  * grad;
            vy += wt * (1 - dx) *    -  1  *      dz  * grad;
            vz += wt * (1 - dx) * (1 - dy) *       1  * grad;
        }
        if (0 <= x + 1 && x + 1 < n1 && 0 <= y     && y     < n2 && 0 <= z + 1 && z + 1 < n3) {
            float grad = dvol[((x + 1) * n2 + y    ) * n3 + z + 1];
            vx += wt *       1  * (1 - dy) *      dz  * grad;
            vy += wt *      dx  *    -  1  *      dz  * grad;
            vz += wt *      dx  * (1 - dy) *       1  * grad;
        }
        if (0 <= x     && x     < n1 && 0 <= y + 1 && y + 1 < n2 && 0 <= z + 1 && z + 1 < n3) {
            float grad = dvol[( x      * n2 + y + 1) * n3 + z + 1];
            vx += wt *    -  1  *      dy  *      dz  * grad;
            vy += wt * (1 - dx) *       1  *      dz  * grad;
            vz += wt * (1 - dx) *      dy  *       1  * grad;
        }
        if (0 <= x + 1 && x + 1 < n1 && 0 <= y + 1 && y + 1 < n2 && 0 <= z + 1 && z + 1 < n3) {
            float grad = dvol[((x + 1) * n2 + y + 1) * n3 + z + 1];
            vx += wt *       1  *      dy  *      dz  * grad;
            vy += wt *      dx  *       1  *      dz  * grad;
            vz += wt *      dx  *      dy  *       1  * grad;
        }
        vs[tid * 3    ] = vx;
        vs[tid * 3 + 1] = vy;
        vs[tid * 3 + 2] = vz;
    }
}
''', 'grad_gridding3d')

class GriddingFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, rho : torch.Tensor, xs : torch.Tensor, shape : Tuple[int]) -> torch.Tensor:
        ctx.rho = rho
        ctx.xs = xs
        rho = torch_to_cupy(rho)
        xs = torch_to_cupy(xs)

        m = len(rho)
        pixel_size = 1 / max(shape)
        assert xs.shape == (m, 3) and rho.shape == (m, ) and len(shape) == 3
        assert rho.device == xs.device

        with rho.device:
            vol = cp.zeros(shape, dtype = cp.float32)
            ker_gridding3d((BLOCKDIM(m), ), (BLOCKSIZE, ), (rho, xs, m, vol, shape[0], shape[1], shape[2], cp.float32(pixel_size)))
            return cupy_to_torch(vol)

    @staticmethod
    def backward(ctx, grad_output : torch.Tensor) -> Tuple[torch.Tensor]:
        rho = torch_to_cupy(ctx.rho)
        xs = torch_to_cupy(ctx.xs)
        grad_output = torch_to_cupy(grad_output)

        shape = grad_output.shape
        m = len(rho)
        pixel_size = 1 / max(shape)
        assert xs.shape == (m, 3) and rho.shape == (m, ) and len(shape) == 3
        assert rho.device == xs.device == grad_output.device

        with rho.device:
            vs = cp.empty_like(xs, dtype = cp.float32)
            ker_grad_gridding3d((BLOCKDIM(m), ), (BLOCKSIZE, ), (rho, xs, m, grad_output, shape[0], shape[1], shape[2], cp.float32(pixel_size), vs))
            return None, cupy_to_torch(vs), None

class Gridding(torch.nn.Module):

    def __init__(self, shape : Tuple[int]):
        super().__init__()
        self.shape = shape

    def forward(self, rho : torch.Tensor, xs : torch.Tensor) -> torch.Tensor:
        return GriddingFunction.apply(rho, xs, self.shape)



__all__ = [
    'particles'
]

import numpy as np
from typing import Tuple
from numpy.typing import NDArray

def particles(vol : NDArray[np.float32], threshold : float = 1e-6) -> Tuple[NDArray[np.float32]]:
    n = vol.shape[0]
    vol = np.where(vol < threshold, 0, vol)
    xs = np.nonzero(vol)
    rho = vol[xs]
    xs = np.vstack(xs)
    xs = (xs.transpose().copy() + 0.5) / n
    return rho, xs



__all__ = [
    'tetramesh',
    'MeshLoss'
]

import numpy as np
import torch
from numpy.typing import NDArray

def tetramesh(xs : NDArray[np.float32], n : NDArray[np.float32]) -> NDArray[np.float32]:
    vertices = np.array([
        [0, 0, 0],
        [0, 1, 0],
        [1, 0, 0],
        [1, 1, 0],
        [0, 0, 1],
        [0, 1, 1],
        [1, 0, 1],
        [1, 1, 1]
    ]) / n
    tetras = vertices[np.array([
        [0, 4, 6, 7],
        [0, 4, 5, 7],
        [0, 1, 5, 7],
        [0, 1, 3, 7],
        [0, 2, 3, 7],
        [0, 2, 6, 7]
    ], dtype = np.int32)]
    return np.reshape(xs[:, None, None] + tetras, (-1, 4, 3))

class MeshLoss(torch.nn.Module):

    def __init__(self, mesh : torch.Tensor):
        super().__init__()
        self.n = len(mesh)
        self.mesh = mesh
        self.inv = torch.linalg.inv(mesh[..., 1:4, :] - mesh[..., 0:1, :])

    def forward(self, odenet : torch.nn.Module) -> torch.Tensor:
        mesh_out = odenet(self.mesh)
        afftrans = self.inv @ (mesh_out[..., 1:4, :] - mesh_out[..., 0:1, :])
        svdvals = torch.linalg.svdvals(afftrans)
        return torch.sum(torch.square(svdvals ** 2 - 1)) / self.n



__all__ = [
    'VelocityNet'
]

import torch

class VelocityNet(torch.nn.Module):

    def __init__(self, depth : int = 3, width : int = 100, dim : int = 3):
        super().__init__()
        self.depth = depth
        for i in range(depth):
            layer = torch.nn.Linear(dim if i == 0 else width, dim if i == depth - 1 else width)
            torch.nn.init.normal_(layer.weight, std = 0.01)
            torch.nn.init.zeros_(layer.bias)
            self.__setattr__(f'n{i}', layer)

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        for i in range(self.depth):
            x = self.__getattr__(f'n{i}')(x)
            if i < self.depth - 1:
                x = torch.nn.functional.leaky_relu(x)
        return x



import argparse
import mrcfile
import sys
import torch
import numpy as np
from numpy.typing import NDArray
from pathlib import Path
from .ode_net import ODENet
from .particles import particles
from .gridding import Gridding

def parse_args():
    parser = argparse.ArgumentParser(description = 'CryoTRANS: Predicting high-resolution maps of rare conformations using neural ODEs in cryo-EM.')

    basic_group = parser.add_argument_group('Basic arguments.')
    basic_group.add_argument('-i', '--initial-map',       type = str,                      help = 'Path of initial map.')
    basic_group.add_argument('-t', '--initial-threshold', type = float, default = 0.,      help = 'Threshold for the initial map.')
    basic_group.add_argument('-d', '--directory',         type = str,                      help = 'Working directory.')
    basic_group.add_argument('-p', '--prefix',            type = str,   default = 'frame', help = 'Prefix for output movie.')
    basic_group.add_argument('-g', '--gpu',               type = int,   default = 0,       help = 'Which gpu to use, 0 by default.')
    basic_group.add_argument('-w', '--weight',            type = str,                      help = 'Path of network weight file as initial model.')

    advanced_group = parser.add_argument_group('Advanced arguments.')
    advanced_group.add_argument('--depth',   type = int,   default = 3,    help = 'Depth of velocity net (MLP).')
    advanced_group.add_argument('--width',   type = int,   default = 100,  help = 'Width of velocity net (MLP).')

    if len(sys.argv) == 1:
        parser.print_help()
        exit()
    return parser.parse_args()

def predict(
    a0 : NDArray[np.float32],
    voxel_size : float,
    odenet : ODENet,
    directory : Path,
    prefix : str
):
    print('------------------------Predicting-------------------------')

    rho, xs = particles(a0)
    rho = torch.tensor(rho, dtype = torch.float32, device = 'cuda')
    xs  = torch.tensor(xs , dtype = torch.float32, device = 'cuda')
    grd = Gridding(a0.shape)

    with torch.no_grad():
        for i, xi in enumerate(odenet.trajectory(xs)):
            print(f'Processing frame {i:02d} ...', end = ' ')
            vol = grd(rho, xi).to('cpu').numpy()
            with mrcfile.new(directory.joinpath(f'{prefix}_{i:02d}.mrc'), data = vol, overwrite = True) as mrc:
                mrc.voxel_size = voxel_size
            print('done.')

def main():
    args = parse_args()

    with mrcfile.open(args.initial_map, permissive = True) as mrc:
        a0 : NDArray[np.float32] = mrc.data
        voxel_size = mrc.voxel_size
        print(f'Voxelsize read from input map: {voxel_size} Angstrom.')
    n = a0.shape[0]
    assert a0.dtype == np.float32
    assert a0.shape == (n, n, n)
    a0 = np.where(a0 < args.initial_threshold, 0, a0)

    directory = Path(args.directory).absolute()
    directory.mkdir(parents = True, exist_ok = True)
    if not directory.is_dir():
        raise RuntimeError(f'Invalid working directory: {args.directory}.')

    torch.cuda.set_device(torch.device('cuda', args.gpu))
    odenet = ODENet(depth = args.depth, width = args.width)
    print(f'Try loading input weight file {args.weight} ...', end = ' ')
    odenet.load_state_dict(torch.load(args.weight))
    print('succeeded!')
    odenet.to('cuda')

    predict(a0, voxel_size, odenet, directory, args.prefix)



import argparse
import mrcfile
import sys
import torch
import numpy as np
from pathlib import Path
from skimage.transform import downscale_local_mean
from time import time
from typing import Optional
from numpy.typing import NDArray
from .ode_net import ODENet
from .particles import particles
from .gridding import Gridding
from .wasserstein2_loss import Wasserstein2Loss
from .tetramesh import tetramesh, MeshLoss

def parse_args():
    parser = argparse.ArgumentParser(description = 'CryoTRANS: Predicting high-resolution maps of rare conformations using neural ODEs in cryo-EM.')

    basic_group = parser.add_argument_group('Basic arguments.')
    basic_group.add_argument('-i0', '--initial-map',       type = str,                     help = 'Path of initial map.')
    basic_group.add_argument('-t0', '--initial-threshold', type = float,                   help = 'Threshold for the initial map.')
    basic_group.add_argument('-i1', '--target-map',        type = str,                     help = 'Path of target map.')
    basic_group.add_argument('-t1', '--target-threshold',  type = float,                   help = 'Threshold for the target map.')
    basic_group.add_argument('-d',  '--directory',         type = str,                     help = 'Working directory.')
    basic_group.add_argument('-g',  '--gpu',               type = int,   default = 0,      help = 'Which gpu to use, 0 by default.')
    basic_group.add_argument('-w',  '--weight',            type = str,   required = False, help = 'Path of network weight file as initial model.')
    basic_group.add_argument('-b',  '--binning',           type = int,   default = 1,      help = 'Binning level, 1 by default.')
    basic_group.add_argument('-n',  '--n_steps',           type = int,                     help = 'Number of training steps.')
    basic_group.add_argument('-p',  '--period',            type = int,                     help = 'For periodic report.')

    advanced_group = parser.add_argument_group('Advanced arguments.')
    advanced_group.add_argument('--depth',   type = int,   default = 3,    help = 'Depth of velocity net (MLP).')
    advanced_group.add_argument('--width',   type = int,   default = 100,  help = 'Width of velocity net (MLP).')
    advanced_group.add_argument('--w2_eps',  type = float, default = 1e-4, help = 'Entropic regularisation parameter for W2 loss.')
    advanced_group.add_argument('--w2_iter', type = int,   default = 5,    help = 'Number of Sinkhorn iteration for computing W2 loss.')
    advanced_group.add_argument('--l2',      action = 'store_true',        help = 'Use L2 refine instead of W2.')
    advanced_group.add_argument('--lr',      type = float, default = 1e-3, help = 'Learning rate. Suggest 1e-4 when doing L2 refine.')
    advanced_group.add_argument('--mu_mesh', type = float,                 help = 'Regularization parameter for tetrahedral mesh loss.')

    if len(sys.argv) == 1:
        parser.print_help()
        exit()
    return parser.parse_args()

def train(
    a0 : NDArray[np.float32],
    a1 : NDArray[np.float32],
    odenet : ODENet,
    n_steps : int,
    period : int,
    directory : Path,
    w2_eps : float = 1e-4,
    w2_iter : int = 5,
    l2 : bool = False,
    lr : float = 1e-3,
    mu_mesh : Optional[float] = None,
):
    rho, xs = particles(a0)
    if mu_mesh is not None:
        mesh = tetramesh(xs, a0.shape[0])
        mesh = torch.tensor(mesh, dtype = torch.float32, device = 'cuda')
        mesh = MeshLoss(mesh)
    rho = torch.tensor(rho, dtype = torch.float32, device = 'cuda')
    xs = torch.tensor(xs, dtype = torch.float32, device = 'cuda')
    a1 = torch.tensor(a1, dtype = torch.float32, device = 'cuda')

    grd = Gridding(a0.shape)
    if not l2:
        w2 = Wasserstein2Loss(eps = w2_eps, maxiter = w2_iter)
    optimizer = torch.optim.Adam(odenet.parameters(), lr = lr)

    losses = []
    loss_file = directory.joinpath('loss.log')
    loss_file.write_text('')

    time0 = time()
    print('|-----------------------------------------|')
    print('|   iter   |       loss       |   time    |')
    print('|----------|------------------|-----------|')

    i_period = 0
    for i_step in range(n_steps + 1):
        optimizer.zero_grad()
        b1 = grd(rho, odenet(xs))
        loss = torch.sum(torch.square(b1 - a1)) if l2 else w2(b1, a1)
        if mu_mesh is not None:
            loss += mu_mesh * mesh(odenet)

        losses.append(loss.item())
        if i_step % period == 0:
            print(f'|{i_step:^10d}|{loss.item():^18.9e}|{time() - time0:^11.3e}|')
            with loss_file.open('+a') as loss_out:
                np.savetxt(loss_out, losses)
                losses.clear()
            torch.save(odenet.state_dict(), directory.joinpath(f'net_{i_period}.pt'))
            i_period += 1

        if i_step != n_steps:
            loss.backward()
            optimizer.step()

    print('|-----------------------------------------|')
    print('Save network ...', end = ' ')
    torch.save(odenet.state_dict(), directory.joinpath('net.pt'))
    print('done.')

def main():
    args = parse_args()

    a0 : NDArray[np.float32] = mrcfile.read(args.initial_map)
    a1 : NDArray[np.float32] = mrcfile.read(args.target_map)
    n = a0.shape[0]
    assert a0.dtype == a1.dtype == np.float32
    assert a0.shape == a1.shape == (n, n, n)

    t0 = args.initial_threshold
    t1 = args.target_threshold
    a0 = np.where(a0 < t0, 0, a0)
    a1 = np.where(a1 < t1, 0, a1)
    a0 /= a0.max()
    a1 *= a0.sum() / a1.sum()

    directory = Path(args.directory).absolute()
    directory.mkdir(parents = True, exist_ok = True)
    if not directory.is_dir():
        raise RuntimeError(f'Invalid working directory: {args.directory}.')

    torch.cuda.set_device(torch.device('cuda', args.gpu))
    odenet = ODENet(depth = args.depth, width = args.width)
    if args.weight is not None:
        try:
            print(f'Try loading input weight file {args.weight} ...', end = ' ')
            odenet.load_state_dict(torch.load(args.weight))
            print('succeeded!')
        except:
            print('failed!')
            print('Random initialization applied.')
    odenet.to('cuda')

    binning = args.binning
    if binning <= 0:
        raise ValueError(f'Binning level should be a positive integer.')
    if n % binning != 0:
        raise ValueError(f'Binning level {binning} dost not divide boxsize.')
    if binning > 1:
        print('Bin maps ...', end = ' ')
        a0 = downscale_local_mean(a0, (binning, binning, binning))
        a1 = downscale_local_mean(a1, (binning, binning, binning))
        print('done.')

    train(
        a0, a1, odenet, args.n_steps, args.period, directory,
        w2_eps = args.w2_eps,
        w2_iter = args.w2_iter,
        l2 = args.l2,
        lr = args.lr,
        mu_mesh = args.mu_mesh
    )



import cupy as cp
import torch
from torch.utils.dlpack import to_dlpack, from_dlpack

BLOCKSIZE = 1024
BLOCKDIM = lambda x : (x - 1) // BLOCKSIZE + 1

def cupy_to_torch(x : cp.ndarray) -> torch.Tensor:
    return from_dlpack(x.toDlpack())

def torch_to_cupy(x : torch.Tensor) -> cp.ndarray:
    return cp.fromDlpack(to_dlpack(x))



__all__ = [
    'Wasserstein2Loss'
]

import cupy as cp
import torch
from typing import Dict, Tuple
from .utility import torch_to_cupy, cupy_to_torch, BLOCKDIM, BLOCKSIZE

ker_blur = cp.RawKernel(r'''
extern "C" __global__ void blur(
    const float* src,
    int size,
    int n,
    float B,
    float* dst)
{
    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    if (tid < size) {
        int offset = tid % n;
        int base = tid - offset;
        float w = src[tid];
        for (int i = 0; i < n; ++i)
            w = fmaxf(w, src[base + i] - (i - offset) * (i - offset) / B);
        float v = 0;
        for (int i = 0; i < n; ++i)
            v += expf(src[base + i] - (i - offset) * (i - offset) / B - w);
        dst[tid] = logf(v) + w;
    }
}
''', 'blur')

def _blur(b, x):
    tmp = cp.empty_like(x, dtype = cp.float32)
    for i in range(x.ndim):
        # move i-th axis to last.
        for j in range(i, x.ndim - 1): x = x.swapaxes(j, j + 1).copy()
        tmp = tmp.reshape(x.shape)

        # blur along last axis.
        ker_blur((BLOCKDIM(x.size), ), (BLOCKSIZE, ), (x, x.size, x.shape[-1], cp.float32(b), tmp))
        x, tmp = tmp, x

        # move last axis back to i.
        for j in range(x.ndim - 1, i, -1): x = x.swapaxes(j - 1, j).copy()
    return x

class Wasserstein2LossFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, a0 : torch.Tensor, a1 : torch.Tensor, paras : Dict) -> torch.Tensor:
        assert torch.is_tensor(a0) and torch.is_tensor(a1)
        assert a0.shape == a1.shape
        assert a0.is_cuda and a1.is_cuda and a0.device == a1.device

        eps = paras.get('eps', 1e-3)
        delta = paras.get('delta', 1e-3)
        maxiter = paras.get('maxiter', 1000)
        batch = paras.get('batch', 100)
        pixel_size = 1 / max(a0.shape)

        a0 = torch_to_cupy(a0)
        a1 = torch_to_cupy(a1)
        with a0.device:
            b = eps / (pixel_size * pixel_size)
            la0 = cp.log(cp.fmax(a0, 1e-20))
            la1 = cp.log(cp.fmax(a1, 1e-20))
            phi0 = cp.zeros_like(a0, dtype = cp.float32)
            phi1 = cp.zeros_like(a1, dtype = cp.float32)
            norm_1 = lambda x : cp.sum(cp.abs(x)).item()
            na0 = norm_1(a0)

            for rd in range(maxiter):
                phi0 = la0 - _blur(b, phi1)
                phi1 = la1 - _blur(b, phi0)
                if rd == 0 or (rd + 1) % batch == 0:
                    err = norm_1(a0 - cp.exp(phi0 + _blur(b, phi1))) / na0
                    # print(f'  Round {rd + 1}, |i0 - P1|_1 / |i0|_1 = {err:.6f}.')
                    if err < delta: break

            loss = -eps * cp.sum(cp.exp(phi0 + _blur(b, phi1)))
            phi0 *= eps
            phi1 *= eps
            loss += cp.vdot(phi0, a0) + cp.vdot(phi1, a1)

        ctx.phi0 = cupy_to_torch(phi0)
        ctx.phi1 = cupy_to_torch(phi1)
        return cupy_to_torch(loss)

    @staticmethod
    def backward(ctx, grad_output : torch.Tensor) -> Tuple:
        return grad_output * ctx.phi0, grad_output * ctx.phi1, None

class Wasserstein2Loss(torch.nn.Module):

    def __init__(self, **paras):
        super().__init__()
        self.paras = paras

    def forward(self, a : torch.Tensor, b : torch.Tensor) -> torch.Tensor:
        return Wasserstein2LossFunction.apply(a, b, self.paras)
